import json
from datetime import datetime

from sqlalchemy import Table, Column, UniqueConstraint, CheckConstraint, event, func, case,  select
from sqlalchemy import ForeignKey, String, Integer, BigInteger
from sqlalchemy.orm import Mapped, mapped_column, relationship, column_property, object_session
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.types import ARRAY, JSON, TypeDecorator, Unicode

from . import Base
from .._types import DatumTypes, ChainTypes

blocks_datums = Table(
    "blocks_datums",
    Base.metadata,
    Column("block_id", ForeignKey("blocks.id"), primary_key=True),
    Column("datum_id", ForeignKey("datums.id"), primary_key=True),
)

events_datums = Table(
    "events_datums",
    Base.metadata,
    Column("event_id", ForeignKey("events.id"), primary_key=True),
    Column("datum_id", ForeignKey("datums.id"), primary_key=True),
)


class DictionaryType(TypeDecorator):
    impl = Unicode

    def process_bind_param(self, value, dialect):
        if isinstance(value, dict):
            return {key: self.process_bind_param(value, dialect) for key, value in value.items()}
        elif isinstance(value, list):
            return [self.process_bind_param(item, dialect) for item in value]
        if not isinstance(value, str):
            value = str(value)
        return value

    def process_result_value(self, value, dialect):
        return json.loads(value)


class DatumBlockAssociation(Base):
    __table__ = blocks_datums


class DatumEventAssociation(Base):
    __table__ = events_datums


class Net(Base):
    __tablename__ = 'nets'
    id: Mapped[int] = mapped_column(
        BigInteger,
        primary_key=True
    )
    name: Mapped[str] = mapped_column(String, unique=True)
    wallets: Mapped[list['Wallet']] = relationship("Wallet", back_populates="net")
    chains: Mapped[list['Chain']] = relationship("Chain", back_populates="net")


class Block(Base):
    __tablename__ = 'blocks'
    id: Mapped[str] = mapped_column(primary_key=True)  # hash + chain_id

    version: Mapped[int] = mapped_column(Integer)
    cell_id: Mapped[str] = mapped_column(String)
    # TODO: move to Chain
    chainid: Mapped[str] = mapped_column(String)
    hash: Mapped[str] = mapped_column(String, nullable=False)
    created_at: Mapped[datetime]
    signs: Mapped[dict] = mapped_column(JSON)
    meta: Mapped[dict] = mapped_column(DictionaryType)

    # o2m
    chain_id: Mapped[int] = mapped_column(BigInteger, ForeignKey('chains.id'), index=True)
    chain: Mapped['Chain'] = relationship(back_populates="blocks")
    # m2m
    datums: Mapped[list['Datum']] = relationship(secondary=blocks_datums, back_populates="blocks")

    # black-magic: block.datums_ids.extend([datum_id, datum_id])
    block_recs: Mapped[list["DatumBlockAssociation"]] = relationship()
    datums_id: Mapped[list[str]] = association_proxy(
        "block_recs", "id",
        creator=lambda d_id: DatumBlockAssociation(datum_id=d_id)
    )

    __table_args__ = (
        UniqueConstraint('hash', 'chain_id', name='uq_block_hash_and_chain_id'),
    )


# auto set id
@event.listens_for(Block, 'before_insert')
def set_id_before_insert(mapper, connection, target: Block):
    target.id = f"{target.hash}{target.chain_id}"


class Event(Base):
    __tablename__ = 'events'
    id: Mapped[str] = mapped_column(primary_key=True)  # hash + chain_id

    version: Mapped[int] = mapped_column(Integer)
    cell_id: Mapped[str] = mapped_column(String)
    hash: Mapped[str] = mapped_column(String, nullable=False)
    created_at: Mapped[datetime]
    signs: Mapped[dict] = mapped_column(JSON)

    chain_id: Mapped[int] = mapped_column(BigInteger, ForeignKey('chains.id'), index=True)
    chain: Mapped['Chain'] = relationship(back_populates="events")
    # m2m
    datums: Mapped[list['Datum']] = relationship(secondary=events_datums, back_populates="events")

    # black-magic: event.datums_ids.extend([datum_id, datum_id])
    event_recs: Mapped[list["DatumEventAssociation"]] = relationship()
    datums_id: Mapped[list[str]] = association_proxy(
        "event_recs", "id",
        creator=lambda d_id: DatumEventAssociation(datum_id=d_id)
    )

    # only events fields:
    round_id: Mapped[int] = mapped_column(Integer)
    hashes: Mapped[list[str]] = mapped_column(ARRAY(String))

    __table_args__ = (
        UniqueConstraint('hash', 'chain_id', name='uq_event_hash_and_chain_id'),
    )


# auto set id
@event.listens_for(Event, 'before_insert')
def set_id_before_insert(mapper, connection, target: Event):
    target.id = f"{target.hash}{target.chain_id}"


class Chain(Base):
    __tablename__ = 'chains'
    id: Mapped[int] = mapped_column(
        BigInteger,
        primary_key=True
    )
    name: Mapped[str] = mapped_column(String)  # not unique
    type: Mapped[ChainTypes]
    net_id: Mapped[int] = mapped_column(BigInteger, ForeignKey('nets.id'))
    net: Mapped['Net'] = relationship(back_populates="chains")
    datums: Mapped[list['Datum']] = relationship(back_populates="chain")
    blocks: Mapped[list['Block']] = relationship(
        back_populates="chain",
        lazy="select",
        # order_by='desc(Block.created_at)'
    )
    events: Mapped[list['Event']] = relationship(
        back_populates="chain",
        lazy="select",
        # order_by='desc(Block.created_at)'
    )

    @property
    def count(self):
        if self.type == ChainTypes.esbocs:
            query = select(func.count(Block.id)).where(Block.chain_id == self.id)
        elif self.type == ChainTypes.dag_poa:
            query = select(func.count(Event.id)).where(Event.chain_id == self.id)
        else:
            raise TypeError

        return object_session(self).execute(query).scalar()

    __table_args__ = (
        UniqueConstraint('name', 'net_id', name='uq_constraint_name_and_net_id'),
    )


class Datum(Base):
    # TODO: обавить счетчик для Атомов и Датумов.
    #  Завесчти счетчик  который считает датумы \ атомы при их чтении из чейна
    __tablename__ = 'datums'
    id: Mapped[str] = mapped_column(String, primary_key=True)  # hash + chain_id
    size: Mapped[int] = mapped_column(Integer)
    version: Mapped[str] = mapped_column(String)
    hash: Mapped[str] = mapped_column(String, nullable=False)
    type: Mapped[DatumTypes]
    created_at: Mapped[datetime]

    chain_id: Mapped[int] = mapped_column(BigInteger, ForeignKey('chains.id'))
    chain: Mapped['Chain'] = relationship(back_populates="datums")

    # block_id: Mapped[str | None] = mapped_column(ForeignKey('blocks.id'))
    blocks: Mapped[list['Block']] = relationship(secondary=blocks_datums, back_populates="datums")

    # event_id: Mapped[str | None] = mapped_column(ForeignKey('events.id'))
    events: Mapped[list['Event']] = relationship(secondary=events_datums, back_populates="datums")

    sub_datum: Mapped[dict] = mapped_column(JSON)

    __table_args__ = (
        UniqueConstraint('hash', 'chain_id', name='uq_datum_hash_and_chain_id'),
        # CheckConstraint('(block_id IS NOT NULL) OR (event_id IS NOT NULL)', name='check_related_atom'),
    )


@event.listens_for(Datum, 'before_insert')
def set_id_before_insert(mapper, connection, target: Datum):
    target.id = f"{target.hash}{target.chain_id}"


class Wallet(Base):
    __tablename__ = 'wallets'
    id: Mapped[str] = mapped_column(String, primary_key=True)  # addr + net_id
    addr: Mapped[str]  # addr - The address is unique, excluding "null" addresses
    net_id: Mapped[int] = mapped_column(BigInteger, ForeignKey('nets.id'))
    net: Mapped['Net'] = relationship("Net", back_populates="wallets")
    tokens: Mapped[list['TokenBalance']] = relationship(back_populates="wallet")

    __table_args__ = (
        UniqueConstraint('addr', 'net_id', name="uq_addr_net_id"),
    )


@event.listens_for(Wallet, 'before_insert')
def set_id_before_insert(mapper, connection, target: Wallet):
    target.id = f"{target.addr}{target.net_id}"


class TokenBalance(Base):
    __tablename__ = 'token_balances'
    id: Mapped[int] = mapped_column(
        BigInteger,
        primary_key=True
    )
    ticker: Mapped[str]
    balance: Mapped[str]  # datoshi
    wallet_id: Mapped[str] = mapped_column(ForeignKey("wallets.id"))
    wallet: Mapped[Wallet] = relationship('Wallet', back_populates="tokens")

    __table_args__ = (
        UniqueConstraint('ticker', 'wallet_id', name="un_ticker_wallet_id"),
    )

