# from typing import NewType
import json
from typing import Literal, Iterable, Callable
from abc import ABC, abstractmethod

from itertools import chain

from web3 import Web3, HTTPProvider
from web3.types import ChecksumAddress, EventData
from web3.middleware import geth_poa_middleware
from web3.contract import Contract
from web3.contract.contract import ContractFunctions, ContractEvents
from web3.exceptions import TransactionNotFound
from web3.datastructures import AttributeDict
from web3._utils import datatypes
from eth_typing import HexStr, HexAddress

import time

log = None 
def set_logger(logger):
    global log 
    log = logger

from requests.exceptions import ConnectionError

class ContractProvider:

    def __init__(self, 
                        name: str, 
                        provider_urls: list[str], 
                        contract_address: ChecksumAddress, 
                        abi: str):
        
        self._name = name
        self._contract_address = contract_address
        self._urls = provider_urls
        self._abi = json.loads(abi)
        self.net_id =  self.w3.net.version
        
    
    def __str__(self):
        return self.name
    
    @property
    def gas_price(self):
        return self.w3.eth.gas_price
    
    @property
    def contract_address(self) -> ChecksumAddress:
        return self._contract_address

    @property
    def name(self) -> str:
        return self._name

    @property
    def urls(self):
        return self._urls

    @property
    def w3(self) -> Web3:
        while True:
            for url in self.urls:
                web3 = Web3(HTTPProvider(url))
                if web3.is_connected():
                    web3.middleware_onion.inject(geth_poa_middleware, layer=0)
                    return web3
            if log:
                log.notice(f"{self.name} No valid links provided - failed to connect")
            time.sleep(3)
            

    def contract(self) -> Contract:
        contract = self.w3.eth.contract(address=self._contract_address, abi=self._abi)
        return contract

    def events(self, event_names: list[str], from_block: int | Literal["latest"], to_block : int|Literal["latest"]):
        contract = self.contract()
        if log:
            log.notice(f"{self.name}: [{contract.w3.provider}] looking for [{event_names}] events from block {from_block} till {to_block}")
        res = []
        for evt in event_names:
            res.extend(contract.events[evt]().get_logs(fromBlock=from_block, toBlock=to_block))
        return res
    def event_from_tx_logs(self, tx_hash : str | HexStr, event_name : str):
        if log:
            log.notice(f"{self.name}: looking for [{event_name}] events for {tx_hash}")
        receipt = self.w3.eth.get_transaction_receipt(tx_hash)
        evt = self.contract().events[event_name]().process_receipt(receipt)
        if evt:
            return evt[0]
        else:
            return None

    def call_function(self, private_key, func_name, args, nonce=None, gas=None):
        contract = self.contract()
        w3 = contract.w3
        pubkey = w3.eth.account.from_key(private_key)
        address_from = pubkey.address

        tx = {
            'from': address_from,
            'gasPrice': gas or  int(w3.eth.gas_price*1.2),
            'nonce': nonce or   w3.eth.get_transaction_count(address_from)
        }   

        if hasattr(contract.functions, func_name):
            contract_function = getattr(contract.functions, func_name)
            transaction = contract_function(*args).build_transaction(tx)
        else:
            if log:
                log.error(f"{self.name}: [{w3.provider}] Function '{func_name}' not found in the contract.")
        
        signed_txn = w3.eth.account.sign_transaction(transaction, private_key=private_key)
        tx_hash = w3.eth.send_raw_transaction(signed_txn.rawTransaction)
        if log:
            log.notice(f"{self.name}: [{w3.provider}] tx sent [{tx_hash.hex()}]")
        return tx_hash



class ContractEventScannerState:
    def __init__(self, start_block : int, current_block : int, window : int = 100, reader=None, writer=None):
        self.start_block = start_block 
        self.block_window = window
        self.end_block = start_block + window

        if self.end_block > current_block:
            self.end_block = current_block

        self.reader = reader
        self.writer = writer
        

    @staticmethod
    def load(provider: ContractProvider, window: int = 100, reader: Callable[[str, ... ],bytes] | None = None, writer : Callable[[str, bytes,...], None] | None = None):
        current_block = provider.w3.eth.block_number

        if reader:
            last_block = int(reader(provider.name).decode("utf-8"))
        else:
            last_block = current_block

        
        if last_block < 0:
            last_block = current_block

        r = ContractEventScannerState(last_block, current_block=current_block, window=window, reader=reader, writer=writer)
        if log:    
            log.notice(f"{provider.name}: initial blocks from: {r.start_block} to: {r.end_block} current: {current_block}")
        return r
    
    def save_current(self, provider: ContractProvider):
        if not self.writer: return
        self.start_block = provider.w3.eth.block_number
        self.end_block = self.start_block
        self.shift(provider)

    def save(self, provider : ContractProvider):
        if not self.writer: return
        self.writer(provider.name, str(self.end_block).encode("utf-8"))
        if log:
            log.notice(f"[{provider.name}]: Saved re-start block: {self.end_block}")

    def shift(self, provider: ContractProvider):
        current_block = provider.w3.eth.block_number

        self.start_block = self.end_block
        self.end_block = self.start_block + self.block_window
        
        if self.end_block > current_block:
            self.end_block = current_block
        if log:
            log.notice(f"[{provider.name}]: Shift scan window: start={self.start_block} end={self.end_block} curr={current_block}")
        

class ContractEventScanner:

    def __init__(self, contract_provider : ContractProvider, event_names : list[str], state: ContractEventScannerState):
        self.provider = contract_provider
        self.event_names = event_names
        self.state = state
        self.logger = None
    
    def set_logger(logger):
        self.log = logger

    def new_events(self):
        events = self.provider.events(self.event_names, self.state.start_block, self.state.end_block)
        self.state.save(self.provider)
        self.state.shift(self.provider)
        return events
    def reset_blocks_numbers(self):
        self.state.save_current(self.provider)
